from warnings import filterwarnings

from mock import patch
from pytest import raises

from kotti.testing import Dummy
from kotti.testing import DummyRequest

# filter deprecation warnings for code that is still tested...
filterwarnings("ignore", "^send_set_password is deprecated")


# noinspection PyAttributeOutsideInit
class TestSendEmail:
    def setup_method(self, method):
        get_mailer_patcher = patch("kotti.message.get_mailer")
        get_mailer = get_mailer_patcher.start()
        self.mailer = get_mailer.return_value

        self.patchers = (get_mailer_patcher,)

    def teardown_method(self, method):
        for patcher in self.patchers:
            patcher.stop()

    def test_send_email(self, dummy_request):
        from kotti.message import send_email

        send_email(
            dummy_request,
            ['"John Doe" <joedoe@foo.com>'],
            "kotti:templates/email-reset-password.pt",
            {
                "site_title": "My site",
                "user_title": "John Doe",
                "url": "http://foo.com",
            },
        )

        assert self.mailer.send.called
        message = self.mailer.send.call_args[0][0]
        assert ['"John Doe" <joedoe@foo.com>'] == message.recipients
        assert "Reset your password" in message.subject

    def test_send_email_without_template_vars(self, dummy_request):
        from kotti.message import send_email

        with raises(NameError):
            send_email(
                dummy_request,
                ['"John Doe" <joedoe@foo.com>'],
                "kotti:templates/email-reset-password.pt",
            )


class TestSendSetPassword:
    def setup_method(self, method):
        get_settings_patcher = patch("kotti.message.get_settings")
        self.get_settings = get_settings_patcher.start()
        self.get_settings.return_value = {
            "kotti.site_title": "Awesome site",
            "kotti.secret2": "123",
        }

        get_mailer_patcher = patch("kotti.message.get_mailer")
        get_mailer = get_mailer_patcher.start()
        self.mailer = get_mailer.return_value

        self.patchers = (get_settings_patcher, get_mailer_patcher)

    def teardown_method(self, method):
        for patcher in self.patchers:
            patcher.stop()

    def test_email_set_password_basic(self, db_session):
        from kotti.message import email_set_password

        user = Dummy(name="joe", email="joe@bar.com", title="Joe")
        email_set_password(user, DummyRequest())

        assert hasattr(user, "confirm_token")
        assert self.mailer.send.called
        message = self.mailer.send.call_args[0][0]
        assert "Your registration" in message.subject
        assert "Joe" in message.body
        assert "Awesome site" in message.body

    def test_email_set_password_add_query(self, db_session):
        from kotti.message import email_set_password

        user = Dummy(name="joe", email="joe@bar.com", title="Joe")
        email_set_password(user, DummyRequest(), add_query={"another": "param"})

        assert self.mailer.send.called
        message = self.mailer.send.call_args[0][0]
        assert "another=param" in message.body


class TestEmailSetPassword:
    def setup_method(self, method):
        get_settings_patcher = patch("kotti.message.get_settings")
        self.get_settings = get_settings_patcher.start()
        self.get_settings.return_value = {
            "kotti.site_title": "Awesome site",
            "kotti.secret2": "123",
        }

        get_mailer_patcher = patch("kotti.message.get_mailer")
        get_mailer = get_mailer_patcher.start()
        self.mailer = get_mailer.return_value

        self.patchers = (get_settings_patcher, get_mailer_patcher)

    def teardown_method(self, method):
        for patcher in self.patchers:
            patcher.stop()

    def test_email_set_password_basic(self, db_session):
        from kotti.message import email_set_password

        user = Dummy(name="joe", email="joe@bar.com", title="Joe")
        email_set_password(user, DummyRequest())

        assert hasattr(user, "confirm_token")
        assert self.mailer.send.called
        message = self.mailer.send.call_args[0][0]
        assert message.subject.startswith("Your registration")
        assert "Joe" in message.body
        assert "Joe" in message.html
        assert "<p" not in message.body
        assert "<p" in message.html
        assert "Awesome site" in message.body

    def test_email_set_password_other_template(self, db_session):
        from kotti.message import email_set_password

        user = Dummy(name="joe", email="joe@bar.com", title="Joe")
        email_set_password(
            user,
            DummyRequest(),
            template_name="kotti:templates/email-reset-password.pt",
        )

        assert self.mailer.send.called
        message = self.mailer.send.call_args[0][0]
        assert message.subject.startswith("Reset your password")

    def test_email_set_password_add_query(self, db_session):
        from kotti.message import email_set_password

        user = Dummy(name="joe", email="joe@bar.com", title="Joe")
        email_set_password(user, DummyRequest(), add_query={"another": "param"})

        assert self.mailer.send.called
        message = self.mailer.send.call_args[0][0]
        assert "another=param" in message.body



""" This module contains navigation related views. """
from pyramid.view import view_config

from kotti.interfaces import INavigationRoot
from kotti.interfaces import INode
from kotti.resources import get_root
from kotti.views.util import nodes_tree


@view_config(
    name="render_tree_navigation",
    permission="view",
    renderer="kotti:templates/edit/nav-tree.pt",
)
@view_config(
    name="navigate", permission="view", renderer="kotti:templates/edit/nav-tree-view.pt"
)
def render_tree_navigation(context, request):
    """ Renders the navigation view.

    :result: Dictionary passed to the template for rendering.
    :rtype: dict
    """
    tree = nodes_tree(request)
    return {"tree": {"children": [tree]}}


@view_config(
    name="local-navigation", context=INode, renderer="kotti:templates/view/nav-local.pt"
)
def local_navigation(context, request):
    def ch(node):
        return [
            child
            for child in node.values()
            if child.in_navigation and request.has_permission("view", child)
        ]

    parent = context
    children = ch(context)
    if not children and getattr(context, "__parent__", None) is not None:
        parent = context.__parent__
        children = ch(parent)
    if children and parent != get_root() and not INavigationRoot.providedBy(parent):
        return dict(parent=parent, children=children)
    return dict(parent=None)


def includeme_local_navigation(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    # Import is needed in function scope to resolve circular imports caused by
    # compatibility imports in slots.py.
    from kotti.views.slots import assign_slot

    config.scan(__name__)
    assign_slot("local-navigation", "right")


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.scan(__name__)



from kotti import get_settings


def test_url_normalizer(config):
    from kotti.url_normalizer import url_normalizer
    assert url_normalizer("simpleandsafe") == "simpleandsafe"
    assert (
        url_normalizer(" Whitespace and capital Letters  ")
        == "whitespace-and-capital-letters"
    )
    assert url_normalizer(">here's another!") == "heres-another"
    assert url_normalizer(">>>here'!--s yet another!!!") == "here-s-yet-another"
    assert url_normalizer("Doe, Joe") == "doe-joe"
    assert url_normalizer("umläut.doc") == "umlaut.doc"
    assert url_normalizer("ZAŻÓŁĆ GĘŚLĄ JAŹŃ") == "zazolc-gesla-jazn"
    assert url_normalizer("zażółć gęślą jaźń") == "zazolc-gesla-jazn"
    assert url_normalizer("quote-this") == "quote-this"
    assert url_normalizer("quote 'this'!") == "quote-this"
    assert url_normalizer("I'm not a FILE.txt") == "im-not-a-file.txt"
    assert url_normalizer("I'm a big file.TXT") == "im-a-big-file.txt"
    assert url_normalizer("rest `n` peace") == "rest-n-peace"
    assert (len(url_normalizer("aa" * 2000))) == 255
    assert url_normalizer("short-hello-version", max_length=10) == "short"


def test_url_normalizer_map_non_ascii_characters(config):
    from kotti.url_normalizer import url_normalizer
    get_settings()["kotti.url_normalizer.map_non_ascii_characters"] = True
    assert url_normalizer("simpleandsafe") == "simpleandsafe"
    assert (
        url_normalizer(" Whitespace and capital Letters  ")
        == "whitespace-and-capital-letters"
    )
    assert url_normalizer(">here's another!") == "heres-another"
    assert url_normalizer(">>>here'!--s yet another!!!") == "here-s-yet-another"
    assert url_normalizer("Doe, Joe") == "doe-joe"
    assert url_normalizer("umläut.doc") == "umlaut.doc"
    assert url_normalizer("ZAŻÓŁĆ GĘŚLĄ JAŹŃ") == "zazolc-gesla-jazn"
    assert url_normalizer("zażółć gęślą jaźń") == "zazolc-gesla-jazn"
    assert url_normalizer("quote-this") == "quote-this"
    assert url_normalizer("quote 'this'!") == "quote-this"
    assert url_normalizer("I'm not a FILE.txt") == "im-not-a-file.txt"
    assert url_normalizer("I'm a big file.TXT") == "im-a-big-file.txt"
    assert url_normalizer("rest `n` peace") == "rest-n-peace"
    assert (len(url_normalizer("aa" * 2000))) == 255
    assert url_normalizer("short-hello-version", max_length=10) == "short"



import json

from mock import MagicMock
from pyramid.security import ALL_PERMISSIONS


class TestNestedMutationDict:
    def test_dictwrapper_comparison(self):
        from kotti.sqla import NestedMutationDict

        assert NestedMutationDict({}) == NestedMutationDict({})
        assert NestedMutationDict({"a": "ok"}) == NestedMutationDict({"a": "ok"})

    def test_listwrapper_comparison(self):
        from kotti.sqla import NestedMutationList

        assert NestedMutationList({}) == NestedMutationList({})
        assert NestedMutationList(["ok"]) == NestedMutationList(["ok"])

    def test_dictwrapper_changed(self):
        from kotti.sqla import NestedMutationDict

        data = {}
        wrapper = NestedMutationDict(data)
        changed = wrapper.changed = MagicMock()

        wrapper["name"] = "andy"
        assert data == {"name": "andy"}
        assert wrapper == {"name": "andy"}
        assert wrapper["name"] == "andy"
        assert changed.call_count == 1

        wrapper["age"] = 77
        assert data == {"name": "andy", "age": 77}
        assert wrapper["age"] == 77
        assert wrapper["name"] == "andy"
        assert changed.call_count == 2

        wrapper["age"] += 1
        assert data == {"name": "andy", "age": 78}
        assert wrapper["age"] == 78
        assert changed.call_count == 3

    def test_listwrapper_changed(self):
        from kotti.sqla import NestedMutationList

        data = []
        wrapper = NestedMutationList(data)
        changed = wrapper.changed = MagicMock()

        wrapper.append(5)
        assert data == [5]
        assert wrapper == [5]
        assert wrapper[0] == 5
        assert changed.call_count == 1

        wrapper.insert(0, 33)
        assert data == [33, 5]
        assert wrapper[0] == 33
        assert changed.call_count == 2

        del wrapper[0]
        assert data == [5]
        assert wrapper[0] == 5
        assert changed.call_count == 3

    def test_dictwrapper_wraps(self):
        from kotti.sqla import NestedMutationDict
        from kotti.sqla import NestedMutationList

        wrapper = NestedMutationDict({"name": "andy", "age": 77, "children": []})
        assert getattr(wrapper, "get", False) is not False
        changed = wrapper.changed = MagicMock()

        wrapper["name"] = "randy"
        assert changed.call_count == 1

        assert isinstance(wrapper["children"], NestedMutationList)
        wrapper["children"].append({"name": "sandy", "age": 33})
        assert changed.call_count == 2
        assert len(wrapper["children"]), 1
        assert isinstance(wrapper["children"][0], NestedMutationDict)

    def test_listwrapper_wraps(self):
        from kotti.sqla import NestedMutationDict
        from kotti.sqla import NestedMutationList

        wrapper = NestedMutationList([{"name": "andy", "age": 77, "children": []}])
        assert getattr(wrapper, "get", False) is False
        changed = wrapper.changed = MagicMock()

        assert isinstance(wrapper[0], NestedMutationDict)
        assert isinstance(wrapper[0]["children"], NestedMutationList)
        assert changed.call_count == 0

    def test_setdefault_dict(self):
        from kotti.sqla import NestedMutationDict

        mdict = NestedMutationDict({})
        assert isinstance(mdict.setdefault("bar", {}), NestedMutationDict)

    def test_setdefault_list(self):
        from kotti.sqla import NestedMutationDict
        from kotti.sqla import NestedMutationList

        mdict = NestedMutationDict({})
        assert isinstance(mdict.setdefault("bar", []), NestedMutationList)

    def test_setdefault_parent(self):
        from kotti.sqla import NestedMutationDict

        mdict = NestedMutationDict({})
        assert mdict.setdefault("bar", []).__parent__ is mdict

    def test_dunder_json(self):
        from kotti.sqla import NestedMutationDict

        data = {"some": ["other", {"stuff": 1}]}
        mdict = NestedMutationDict(data)

        assert json.loads(json.dumps(mdict.__json__(None))) == data


class TestJsonType:
    def make(self):
        from kotti.sqla import JsonType

        return JsonType()

    def test_process_bind_param_no_value(self):
        value = self.make().process_bind_param(None, None)
        assert value is None

    def test_process_bind_param_with_value(self):
        value = self.make().process_bind_param([{"foo": "bar"}], None)
        assert value == '[{"foo": "bar"}]'

    def test_process_bind_param_with_mutationlist(self):
        from kotti.sqla import MutationList

        value = self.make().process_bind_param(MutationList([{"foo": "bar"}]), None)

        assert value == '[{"foo": "bar"}]'

    def test_process_result_value_no_value(self):
        value = self.make().process_result_value(None, None)
        assert value is None

    def test_process_result_value_with_value(self):
        value = self.make().process_result_value('[{"foo": "bar"}]', None)
        assert value == [{"foo": "bar"}]


class TestACLType:
    def make(self):
        from kotti.sqla import ACLType

        return ACLType()

    def test_process_bind_param_no_value(self):
        value = self.make().process_bind_param(None, None)
        assert value is None

    def test_process_bind_param_with_value(self):
        value = self.make().process_bind_param([("Allow", "role:admin", "edit")], None)
        assert value == '[["Allow", "role:admin", "edit"]]'

    def test_process_bind_param_with_default_permissions(self):
        acl = [("Allow", "role:admin", ALL_PERMISSIONS)]
        value = self.make().process_bind_param(acl, None)
        assert value == "[]"

    def test_process_bind_param_with_empty_list(self):
        value = self.make().process_bind_param([], None)
        assert value == "[]"

    def test_process_bind_param_with_default_permissions_and_others(self):
        acl = [("Allow", "role:admin", ALL_PERMISSIONS), ("Deny", "role:admin", "edit")]
        value = self.make().process_bind_param(acl, None)
        assert value == '[["Deny", "role:admin", "edit"]]'
        assert self.make().process_result_value(value, None) == acl

    def test_process_result_value_no_value(self):
        value = self.make().process_result_value(None, None)
        assert value is None

    def test_process_result_value_with_value(self):
        acl = self.make().process_result_value(
            '[["Allow", "role:admin", "edit"]]', None
        )
        assert acl == [
            ("Allow", "role:admin", ALL_PERMISSIONS),
            ("Allow", "role:admin", "edit"),
        ]


class TestMutationList:
    def test_radd(self):
        from kotti.sqla import MutationList

        mlist = MutationList(["foo"])
        assert ["bar"] + mlist == ["bar", "foo"]


class TestMutationDunderJson:
    def test_dunder_json(self):
        from kotti.sqla import MutationList

        mlist = MutationList(["foo"])
        json.loads(json.dumps(mlist.__json__())) == ["foo"]

    def test_dunder_json_recursive(self):
        from kotti.sqla import MutationList
        from kotti.sqla import MutationDict

        mlist = MutationList(
            [
                MutationDict({"foo": MutationList([{"bar": "baz"}])}),
                {"foo": ["bar", "baz"]},
            ]
        )

        json.loads(json.dumps(mlist.__json__())) == [
            {"foo": [{"bar": "baz"}]},
            {"foo": ["bar", "baz"]},
        ]

        mdict = MutationDict(
            {"foo": MutationList([{"bar": "baz"}]), "bar": ["bar", "baz"]}
        )

        json.loads(json.dumps(mdict.__json__())) == {
            "foo": [{"bar": "baz"}],
            "bar": ["bar", "baz"],
        }


def test_missing_foreign_key_indices(db_session):
    """ Test for foreign key constraints that don't have a corresponding
    index. """

    from sqlalchemy_utils.functions import non_indexed_foreign_keys
    from kotti import metadata

    assert non_indexed_foreign_keys(metadata) == {}



class TestExtendingRequest:
    def test_it(self):
        from kotti.request import Request
        from zope.interface import providedBy, implementedBy

        req = Request({})
        req.set_property(lambda x: "exists", "marker", reify=True)

        assert providedBy(req) == implementedBy(Request)
        assert req.marker == "exists"



import json

from pytest import raises
from zope.testbrowser.browser import LinkNotFoundError

from kotti.testing import BASE_URL
from kotti.testing import user
from kotti.views.edit.upload import UploadView


def test_upload_anonymous(root, dummy_request, browser):

    view = UploadView(root, dummy_request)

    assert view.factories == []

    link = browser.getLink

    browser.open(f"{BASE_URL}/")

    # There must be no Upload Link for anonymous users
    with raises(LinkNotFoundError):
        link("Upload Content").click()

    # Upload views must redirect to login for anonymous users
    browser.open(f"{BASE_URL}/upload")
    assert browser.url.startswith(f"{BASE_URL}/@@login")

    browser.open(f"{BASE_URL}/content_types")
    assert browser.url.startswith(f"{BASE_URL}/@@login")


@user("admin")
def test_upload_authenticated_wo_mimetype(root, dummy_request, browser):

    # cannot call content_types without mimetype
    with raises(KeyError):
        browser.open(f"{BASE_URL}/content_types")


@user("admin")
def test_upload_authenticated_text(root, dummy_request, browser):

    # get possible content types for text/plain
    browser.open(f"{BASE_URL}/content_types?mimetype=text/plain")
    c = browser.contents
    if isinstance(c, bytes):
        c = c.decode()
    j = json.loads(c)
    assert "content_types" in j

    # only files are allowed
    types = j["content_types"]
    assert len(types) == 1
    assert types[0]["name"] == "File"



import pytest


@pytest.mark.user("admin")
def test_traversal(webtest, root):
    from kotti.resources import Document
    d1 = root["d1"] = Document(title="Title 1", body="Body 1")
    d2 = d1["d2"] = Document(title="Title 2", body="Body 2")
    d2["d3"] = Document(title="Title 3", body="Body 3")
    app = webtest.app

    resp = app.get("/d1")
    assert "Title 1" in resp.text
    assert "Body 1" in resp.text

    resp = app.get("/d1/d2")
    assert "Title 1" in resp.text
    assert "Title 2" in resp.text
    assert "Body 2" in resp.text

    resp = app.get("/d1/d2/d3")
    assert "Title 1" in resp.text
    assert "Title 2" in resp.text
    assert "Title 3" in resp.text
    assert "Body 3" in resp.text



from kotti.testing import DummyRequest


def _create_contents(root):
    from kotti.resources import Content, File

    doc1 = root["doc1"] = Content(title="First Document", description="I am the first.")
    doc11 = root["doc1"]["doc11"] = Content(
        title="Second Document", description="And I am the second."
    )
    doc12 = root["doc1"]["doc12"] = Content(title="Third Document")
    file1 = root["doc1"]["file1"] = File(
        title="First File", description="this is a file"
    )
    return doc1, doc11, doc12, file1


def _create_contents_with_tags(root=None):
    from kotti.resources import get_root
    from kotti.resources import Content, File

    if root is None:
        root = get_root()

    animals = root["animals"] = Content(title="Animals")
    cat = root["animals"]["cat"] = Content(title="Cat")
    dog = root["animals"]["dog"] = Content(title="Dog")
    monkey = root["animals"]["monkey"] = Content(title="Monkey")
    gorilla = root["animals"]["gorilla"] = Content(title="Gorilla")
    monkey_file = root["animals"]["monkey_file"] = File(
        title="Monkey File",
        description="A Rhesus Macaque and a Green Monkey walk into a bar...",
    )

    root["animals"]["cat"].tags = ["Animals", "Cat"]
    root["animals"]["dog"].tags = ["Animals", "Dog"]
    root["animals"]["monkey"].tags = ["Animals", "Monkey", "Primate"]
    root["animals"]["monkey_file"].tags = ["Animals", "Monkey", "Primate"]
    root["animals"]["gorilla"].tags = ["Animals", "Gorilla", "Primate"]

    return animals, cat, dog, monkey, gorilla, monkey_file


def test_search_empty_content(db_session):
    from kotti.views.util import search_content

    request = DummyRequest()
    results = search_content("teststring", request)
    assert results == []


def test_search_content(root):
    from kotti.views.util import search_content
    from kotti import DBSession
    from kotti.resources import Tag

    request = DummyRequest()
    doc1, doc11, doc12, file1 = _create_contents(root)
    results = search_content("First Document", request)
    assert len(results) == 1
    assert results[0]["name"] == "doc1"
    assert results[0]["title"] == "First Document"
    results = search_content("Document", request)
    # The frontpage contains 'Documentation' in its body!
    assert len(results) == 4
    assert results[1]["name"] == "doc11"
    assert results[1]["title"] == "Second Document"
    assert results[1]["path"] == "/doc1/doc11/"
    assert results[1]["path"][-1] == "/"

    animals, cat, dog, monkey, gorilla, monkey_file = _create_contents_with_tags()

    tags = DBSession.query(Tag).all()
    assert len(tags) == 6
    results = search_content("Animals", request)
    assert len(results) == 6
    results = search_content("Cat", request)
    assert len(results) == 1
    results = search_content("Primate", request)
    assert len(results) == 3

    # Tags were included in general search by modifying the pre-existing
    # approach, wherein searching first worked on title and description,
    # then on body, so that the search order became:
    #
    #     first on title + description
    #         then on tags
    #             then on body
    #
    # So we test here to assure search results come back in that order.
    # Searching on 'Animals', we should find all 6 content items, and the
    # first item should be the Animals folder, found with a title hit, and
    # the other items were found via tags.
    #
    # Note: this ordering is done to have some method, but it does not
    #       necessarily constitute a specification.
    #
    results = search_content("Animals", request)
    assert len(results) == 6
    assert results[0]["name"] == "animals"


def test_search_file_description(root):
    from kotti.views.util import search_content

    request = DummyRequest()
    doc1, doc11, doc12, file1 = _create_contents(root)
    results = search_content("this is a file", request)
    assert len(results) == 1
    assert results[0]["name"] == "file1"
    assert results[0]["title"] == "First File"
    assert results[0]["path"] == "/doc1/file1/"


def test_search_content_without_permission(config, root):
    from kotti.views.util import search_content

    request = DummyRequest()
    _create_contents(root)
    config.testing_securitypolicy(permissive=False)
    results = search_content("Document", request)
    assert len(results) == 0


def test_search_functional(webtest, root):

    doc1, doc11, doc12, file1 = _create_contents(root)

    resp = webtest.app.get("/")

    search_form = resp.forms["form-search"]
    search_form["search-term"] = "First Document"
    resp = search_form.submit()
    assert "I am the first" in resp.text
    assert "And I am the second" not in resp.text

    search_form = resp.forms["form-search"]
    search_form["search-term"] = "Document"
    resp = search_form.submit()
    assert "I am the first" in resp.text
    assert "And I am the second" in resp.text

    search_form = resp.forms["form-search"]
    search_form["search-term"] = "is a file"
    resp = search_form.submit()
    assert "And I am the second" not in resp.text
    assert "Third" not in resp.text
    assert "First File" in resp.text



import datetime

import pytest

from kotti.filedepot import DBFileStorage
from kotti.filedepot import DBStoredFile
from kotti.resources import File


class TestDBStoredFile:
    def test_storedfile_interface(self, db_session, events, setup_app):
        f = DBStoredFile(
            "fileid",
            filename="f.jpg",
            content_type="image/jpeg",
            content_length=1000,
            data=b"content",
        )

        assert f.close() is None
        assert f.closed() is False
        assert f.seekable() is True
        assert f.writable() is False

        assert f.read() == b"content"
        assert f.read() == b""
        f.seek(0)
        assert f.read() == b"content"
        f.seek(0)
        assert f.read(-1) == b"content"
        f.seek(0)
        assert f.read(2) == b"co"
        assert f.read(4) == b"nten"
        assert f.tell() == 6
        f.seek(0)
        f.seek(100)
        assert f.tell() == 100
        assert f.read() == b""

        assert f.content_length == 1000
        assert f.content_type == "image/jpeg"
        assert f.file_id == "fileid"
        assert f.filename == "f.jpg"
        assert f.name == "f.jpg"
        assert f.public_url is None

        f.data = None
        db_session.add(f)
        db_session.flush()
        assert f.content_length == 0

    def test_content_length(self, db_session, events, setup_app):
        f = DBStoredFile("fileid", data=b"content")
        db_session.add(f)
        db_session.flush()

        assert f.content_length == 7

        f.data = b"content changed"
        db_session.flush()

        assert f.content_length == len(b"content changed")

    def test_last_modified(self, monkeypatch, db_session, events, setup_app):
        from kotti import filedepot

        now = datetime.datetime.now()

        class mockdatetime:
            @staticmethod
            def now():
                return now

        monkeypatch.setattr(filedepot, "datetime", mockdatetime)

        f = DBStoredFile("fileid", data=b"content")
        db_session.add(f)
        db_session.flush()

        assert f.last_modified == now

        f.last_modified = None
        f.data = b"content changed"
        db_session.flush()

        assert f.last_modified == now


class TestDBFileStorage:
    def make_one(
        self, content=b"content here", filename="f.jpg", content_type="image/jpg"
    ):

        file_id = DBFileStorage().create(
            content=content, filename=filename, content_type=content_type
        )
        return file_id

    def test_create(self, db_session):
        file_id = self.make_one()
        assert len(file_id) == 36

        fs = db_session.query(DBStoredFile).filter_by(file_id=file_id).one()
        assert fs.data == b"content here"

    def test_list(self):
        with pytest.raises(NotImplementedError):
            DBFileStorage().list()

    def test_exists(self, db_session):
        assert DBFileStorage().exists("1") is False
        file_id = self.make_one()
        assert DBFileStorage().exists(file_id) is True

    def test_get(self, db_session):
        with pytest.raises(IOError):
            DBFileStorage().get("1")

        file_id = self.make_one()
        assert DBFileStorage().get(file_id).data == b"content here"

    def test_delete(self, db_session):
        file_id = DBFileStorage().create(b"content here", "f.jpg", "image/jpg")
        fs = DBFileStorage().get(file_id)

        db_session.add(fs)
        db_session.flush()

        assert db_session.query(DBStoredFile.file_id).one()[0] == file_id

        DBFileStorage().delete(file_id)
        assert db_session.query(DBStoredFile).count() == 0

    def test_replace(self, db_session):
        file_id = self.make_one()

        DBFileStorage().replace(file_id, b"second content", "f2.jpg", "doc")
        fs = DBFileStorage().get(file_id)
        assert fs.filename == "f2.jpg"
        assert fs.content_type == "doc"
        assert fs.read() == b"second content"

        DBFileStorage().replace(fs, b"third content", "f3.jpg", "xls")
        assert fs.filename == "f3.jpg"
        assert fs.content_type == "xls"
        assert fs.read() == b"third content"

    def test_session_integration(self, db_session):
        from depot.manager import DepotManager

        DepotManager._default_depot = "default"
        DepotManager._depots = {"default": DBFileStorage()}

        file_id = DepotManager.get().create(b"content here", "f.jpg", "image/jpg")
        fs = DepotManager.get().get(file_id)

        db_session.add(fs)
        import transaction

        transaction.commit()

        transaction.begin()
        db_session.delete(fs)
        transaction.commit()

        with pytest.raises(IOError):
            DepotManager.get().get(file_id)


class TestMigrateBetweenStorage:
    def _create_content(self, db_session, root, image1, image2):
        data = [
            (b"f1...", "file1.jpg", "image/jpeg"),
            (b"f2...", "file2.png", "image/png"),
        ]
        for row in data:
            f = File(data=row[0], filename=row[1], mimetype=row[2])
            root[row[1]] = f

        data = [
            (image2, "image1.jpg", "image/jpeg"),
            (image1, "image2.png", "image/png"),
        ]
        for row in data:
            f = File(data=row[0], filename=row[1], mimetype=row[2])
            root[row[1]] = f

        db_session.flush()

    def test_migrate_between_storages(
        self, db_session, root, no_filedepots, image_asset, image_asset2
    ):
        from kotti.filedepot import configure_filedepot
        from kotti.filedepot import migrate_storage
        from kotti.resources import Node
        from depot.fields.sqlalchemy import _SQLAMutationTracker
        from sqlalchemy import event
        import os
        import tempfile
        import shutil

        event.listen(
            db_session, "before_commit", _SQLAMutationTracker._session_committed
        )

        tmp_location = tempfile.mkdtemp()

        settings = {
            "kotti.depot.0.backend": "kotti.filedepot.DBFileStorage",
            "kotti.depot.0.name": "dbfiles",
            "kotti.depot.1.backend": "depot.io.local.LocalFileStorage",
            "kotti.depot.1.name": "localfs",
            "kotti.depot.1.storage_path": tmp_location,
        }

        configure_filedepot(settings)
        image1 = image_asset.read()
        image2 = image_asset2.read()
        self._create_content(db_session, root, image1, image2)

        assert db_session.query(DBStoredFile).count() == 4

        migrate_storage("dbfiles", "localfs")

        folders = os.listdir(tmp_location)
        assert len(folders) == 4

        db_session.flush()

        # here we need a transaction.commit(), but that would mess with the
        # rest of the tests; we'll just trigger the event handler manually
        _SQLAMutationTracker._session_committed(db_session)

        root = db_session.query(Node).filter_by(parent=None).one()
        f1 = root["file1.jpg"]
        assert f1.data.file_id in folders
        assert f1.data.file.read() == b"f1..."

        f2 = root["file2.png"]
        assert f2.data.file_id in folders
        assert f2.data.file.read() == b"f2..."

        i1 = root["image1.jpg"]
        assert i1.data.file_id in folders
        i1data = i1.data.file.read()
        assert i1data == image2

        i2 = root["image2.png"]
        assert i2.data.file_id in folders
        i2data = i2.data.file.read()
        assert i2data == image1

        assert db_session.query(DBStoredFile).count() == 0

        shutil.rmtree(tmp_location)


class TestTween:
    @pytest.mark.user("admin")
    def test_tween(self, webtest, filedepot, root, image_asset, db_session):

        from kotti.resources import File
        from kotti.resources import get_root

        # create an image resource
        root["img"] = File(data=image_asset.read(), title="Image")
        db_session.flush()
        root = get_root()
        img = root["img"]

        # the image resource itself is served by the full Kotti stack
        resp = webtest.app.get("/img")
        assert resp.content_type == "text/html"
        assert resp.etag is None
        assert resp.cache_control.max_age == 0
        assert '<a href="http://localhost/img/@@attachment-view">' in resp.text

        # test 404
        resp = webtest.app.get("/depot/non_existing/fileid", status=404)
        assert resp.status_code == 404



import colander
import pytest
from mock import Mock
from mock import patch
from pytest import raises

from kotti.testing import DummyRequest


class TestUserManagement:
    def test_roles(self, root):
        from kotti.security import USER_MANAGEMENT_ROLES
        from kotti.views.users import UsersManage

        request = DummyRequest()
        assert [
            r.name for r in UsersManage(root, request)()["available_roles"]
        ] == USER_MANAGEMENT_ROLES

    def test_search(self, extra_principals, root):
        from kotti.security import get_principals
        from kotti.views.users import UsersManage

        request = DummyRequest()
        P = get_principals()

        request.params["search"] = ""
        request.params["query"] = "Joe"
        entries = UsersManage(root, request)()["entries"]
        assert len(entries) == 0
        assert request.session.pop_flash("info") == ["No users or groups were found."]
        request.params["query"] = "Bob"
        entries = UsersManage(root, request)()["entries"]
        assert entries[0][0] == P["bob"]
        assert entries[0][1] == ([], [])
        assert entries[1][0] == P["group:bobsgroup"]
        assert entries[1][1] == ([], [])

        P["bob"].groups = ["group:bobsgroup"]
        P["group:bobsgroup"].groups = ["role:admin"]
        entries = UsersManage(root, request)()["entries"]
        # assert entries[0][1] == (['group:bobsgroup', 'role:admin'], ['role:admin'])  # noqa
        assert set(entries[0][1][0]) == {"group:bobsgroup", "role:admin"}
        assert entries[0][1][1] == ["role:admin"]
        assert entries[1][1] == (["role:admin"], [])

    def test_apply(self, extra_principals, root):
        from kotti.security import get_principals
        from kotti.security import list_groups
        from kotti.views.users import UsersManage

        request = DummyRequest()

        bob = get_principals()["bob"]

        request.params["apply"] = ""
        UsersManage(root, request)()
        assert request.session.pop_flash("info") == ["No changes were made."]
        assert list_groups("bob") == []
        bob.groups = ["role:special"]

        request.params["role::bob::role:owner"] = "1"
        request.params["role::bob::role:editor"] = "1"
        request.params["orig-role::bob::role:owner"] = ""
        request.params["orig-role::bob::role:editor"] = ""

        UsersManage(root, request)()
        assert request.session.pop_flash("success") == ["Your changes have been saved."]
        assert set(list_groups("bob")) == {"role:owner", "role:editor", "role:special"}

    def test_group_validator(self, db_session):
        from kotti.views.users import group_validator

        with raises(colander.Invalid):
            group_validator(None, "this-group-never-exists")


class TestUserDelete:
    def test_user_delete(self, events, extra_principals, root):
        from kotti.security import get_principals
        from kotti.views.users import user_delete

        request = DummyRequest()
        bob = get_principals()["bob"]

        request.params["name"] = ""
        user_delete(root, request)
        assert request.session.pop_flash("error") == ["No name was given."]
        assert "bob" in get_principals().keys()

        request.params["name"] = "bob"
        result = user_delete(root, request)
        assert "api" in result
        api = result["api"]
        assert api.principal == bob
        assert api.principal_type == "User"
        assert "bob" in get_principals().keys()

        request.params["name"] = "john"
        request.params["delete"] = "delete"
        user_delete(root, request)
        assert request.session.pop_flash("error") == ["User was not found."]
        assert "bob" in get_principals().keys()

        request.params["name"] = "bob"
        request.params["delete"] = "delete"
        user_delete(root, request)
        with pytest.raises(KeyError):
            # noinspection PyStatementEffect
            get_principals()["bob"]

    def test_deleted_group_removed_in_usergroups(
        self, events, extra_principals, root, db_session
    ):
        from kotti.security import get_principals
        from kotti.views.users import user_delete

        request = DummyRequest()
        bob = get_principals()["bob"]
        bob.groups = ["group:bobsgroup"]
        assert bob.groups == ["group:bobsgroup"]

        request.params["name"] = "group:bobsgroup"
        request.params["delete"] = "delete"
        user_delete(root, request)
        db_session.expire(bob)
        with pytest.raises(KeyError):
            # noinspection PyStatementEffect
            get_principals()["group:bobsgroup"]
        assert bob.groups == []

    def test_deleted_group_removed_from_localgroups(
        self, events, extra_principals, root
    ):
        from kotti.security import set_groups
        from kotti.resources import LocalGroup
        from kotti.views.users import user_delete

        request = DummyRequest()
        set_groups("group:bobsgroup", root, ["role:admin"])
        local_group = LocalGroup.query.first()
        assert local_group.principal_name == "group:bobsgroup"
        assert local_group.node == root

        request.params["name"] = "group:bobsgroup"
        request.params["delete"] = "delete"
        user_delete(root, request)
        assert LocalGroup.query.first() is None

    def test_reset_owner_to_none(self, events, extra_principals, root):
        from kotti.resources import Content
        from kotti.views.users import user_delete

        request = DummyRequest()

        root["content_1"] = Content()
        root["content_1"].owner = "bob"
        assert root["content_1"].owner == "bob"

        request.params["name"] = "bob"
        request.params["delete"] = "delete"
        user_delete(root, request)
        assert root["content_1"].owner is None


# noinspection PyAttributeOutsideInit
class TestSetPassword:
    def setup_method(self, method):
        Form_patcher = patch("kotti.views.login.Form")
        self.Form_mock = Form_patcher.start()

        _find_user_patcher = patch("kotti.views.login._find_user")
        self._find_user_mock = _find_user_patcher.start()
        self.user = self._find_user_mock.return_value

        validate_token_patcher = patch("kotti.views.login.validate_token")
        self.validate_token_mock = validate_token_patcher.start()

        self.patchers = (Form_patcher, _find_user_patcher, validate_token_patcher)

    def teardown_method(self, method):
        for patcher in self.patchers:
            patcher.stop()

    def form_values(self, values):
        self.Form_mock.return_value.validate.return_value = values

    def test_success(self, root):
        from kotti.security import get_principals
        from kotti.views.login import set_password

        self.form_values(
            {
                "token": "mytoken",
                "email": "myemail",
                "password": "mypassword",
                "continue_to": "",
            }
        )
        self.user.confirm_token = "mytoken"
        self.user.password = "old_password"
        context, request = root, DummyRequest(post={"submit": "submit"})
        self.user.last_login_date = None
        res = set_password(context, request)

        assert self.user.confirm_token is None
        assert self.user.last_login_date is not None
        assert get_principals().validate_password("mypassword", self.user.password)
        assert res.status == "302 Found"

    def test_wrong_token(self, root):
        from kotti.security import get_principals
        from kotti.views.login import set_password

        self.form_values(
            {
                "token": "wrongtoken",
                "email": "myemail",
                "password": "mypassword",
                "continue_to": "",
            }
        )
        self.user.confirm_token = "mytoken"
        self.user.password = "old_password"
        context, request = root, DummyRequest(post={"submit": "submit"})
        res = set_password(context, request)

        assert self.user.confirm_token == "mytoken"
        assert not get_principals().validate_password("mypassword", self.user.password)
        assert not request.is_response(res)

    def test_inactive_user(self, root):
        from kotti.security import get_principals
        from kotti.views.login import set_password

        self.form_values(
            {
                "token": "mytoken",
                "email": "myemail",
                "password": "mypassword",
                "continue_to": "",
            }
        )
        self.user.confirm_token = "mytoken"
        self.user.password = "old_password"
        context, request = root, DummyRequest(post={"submit": "submit"})
        self.user.active = False
        self.user.last_login_date = None
        res = set_password(context, request)

        assert self.user.confirm_token == "mytoken"
        assert self.user.last_login_date is None
        assert not get_principals().validate_password("mypassword", self.user.password)
        assert not request.is_response(res)

    def test_success_continue(self, root):
        from kotti.views.login import set_password

        self.form_values(
            {
                "token": "mytoken",
                "email": "myemail",
                "password": "mypassword",
                "continue_to": "http://example.com/here#there",
            }
        )
        self.user.confirm_token = "mytoken"
        context, request = root, DummyRequest(post={"submit": "submit"})
        res = set_password(context, request)

        assert res.status == "302 Found"
        assert res.location == "http://example.com/here#there"


class TestUserManageForm:
    def test_schema_factory(self, root):
        from kotti.views.users import UserManageFormView

        request = DummyRequest()
        view = UserManageFormView(root, request)

        schema = view.schema_factory()
        assert "name" not in schema
        assert "password" in schema

    def test_form(self, root):
        from kotti.views.users import UserManageFormView

        request = DummyRequest()
        form = UserManageFormView(root, request)()
        assert 'input type="password"' in form["form"]

    def test_hashed_password_save(self, root):
        from kotti.views.users import UserManageFormView

        user = Mock()
        request = DummyRequest()
        view = UserManageFormView(user, request)
        appstruct = {"password": "foo"}
        view.save_success(appstruct)
        assert user.password.startswith("$2b$10$")

    def test_hashed_password_empty(self, root):
        from kotti.views.users import UserManageFormView

        user = Mock(password="before")
        request = DummyRequest()
        view = UserManageFormView(user, request)
        appstruct = {"password": ""}
        view.save_success(appstruct)
        assert user.password == "before"



"""
Edit views.
"""
import warnings

from kotti.util import _
from kotti.views.edit.content import ContentSchema
from kotti.views.edit.content import DocumentSchema
from kotti.workflow import get_workflow

# API
ContentSchema = ContentSchema
DocumentSchema = DocumentSchema


def _translate_titles(info):
    result = []
    for d in info:
        d = d.copy()
        try:
            d["title"] = eval(d["title"]) if "title" in d else d["name"]
            warnings.warn(
                "_() in workflow.zcml is deprecated. "
                "Support will be removed in Kotti 2.0.0.",
                DeprecationWarning,
            )
        except (NameError, SyntaxError):
            d["title"] = _(d["title"]) if "title" in d else d["name"]
        result.append(d)
    return result


def _state_info(context, request):
    wf = get_workflow(context)
    state_info = []
    if wf is not None:
        state_info = _translate_titles(wf.state_info(context, request))
    return state_info


def _states(context, request):
    state_info = _state_info(context, request)
    return {i["name"]: i for i in state_info}


# noinspection PyUnusedLocal
def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    pass



import os.path

import pkg_resources
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.config import Configurator
from pyramid.events import BeforeRender
from pyramid.threadlocal import get_current_registry
from pyramid.util import DottedNameResolver
from pyramid_beaker import session_factory_from_settings
from sqlalchemy import MetaData
from sqlalchemy import engine_from_config
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from zope.sqlalchemy import register

from kotti.sqla import Base as KottiBase

metadata = MetaData()
DBSession = scoped_session(sessionmaker(autoflush=True))
register(DBSession)
Base = declarative_base(cls=KottiBase)
Base.metadata = metadata
Base.query = DBSession.query_property()
TRUE_VALUES = ("1", "y", "yes", "t", "true")
FALSE_VALUES = ("0", "n", "no", "f", "false", "none")


def authtkt_factory(**settings):
    from kotti.security import list_groups_callback

    kwargs = dict(
        secret=settings["kotti.secret2"],
        hashalg="sha512",
        callback=list_groups_callback,
    )
    try:
        return AuthTktAuthenticationPolicy(**kwargs)
    except TypeError:
        # BBB with Pyramid < 1.4
        kwargs.pop("hashalg")
        return AuthTktAuthenticationPolicy(**kwargs)


def acl_factory(**settings):
    return ACLAuthorizationPolicy()


def beaker_session_factory(**settings):
    return session_factory_from_settings(settings)


def none_factory(**kwargs):  # pragma: no cover
    return None


# All of these can be set by passing them in the Paste Deploy settings:
conf_defaults = {
    "kotti.alembic_dirs": "kotti:alembic",
    "kotti.asset_overrides": "",
    "kotti.authn_policy_factory": "kotti.authtkt_factory",
    "kotti.authz_policy_factory": "kotti.acl_factory",
    "kotti.available_types": " ".join(
        ["kotti.resources.Document", "kotti.resources.File"]
    ),
    "kotti.base_includes": " ".join(
        [
            "kotti",
            "kotti.traversal",
            "kotti.filedepot",
            "kotti.events",
            "kotti.sanitizers",
            "kotti.views",
            "kotti.views.cache",
            "kotti.views.view",
            "kotti.views.edit",
            "kotti.views.edit.actions",
            "kotti.views.edit.content",
            "kotti.views.edit.default_views",
            "kotti.views.edit.upload",
            "kotti.views.file",
            "kotti.views.login",
            "kotti.views.navigation",
            "kotti.views.users",
        ]
    ),
    "kotti.caching_policy_chooser": (
        "kotti.views.cache.default_caching_policy_chooser"
    ),
    "kotti.configurators": "",
    "kotti.date_format": "medium",
    "kotti.datetime_format": "medium",
    "kotti.depot_mountpoint": "/depot",
    "kotti.depot_replace_wsgi_file_wrapper": False,
    "kotti.depot.0.backend": "kotti.filedepot.DBFileStorage",
    "kotti.depot.0.name": "dbfiles",
    "kotti.fanstatic.edit_needed": "kotti.fanstatic.edit_needed",
    "kotti.fanstatic.view_needed": "kotti.fanstatic.view_needed",
    "kotti.login_success_callback": "kotti.views.login.login_success_callback",
    "kotti.max_file_size": "10",
    "kotti.modification_date_excludes": " ".join(["kotti.resources.Node.position"]),
    "kotti.populators": "kotti.populate.populate",
    "kotti.principals_factory": "kotti.security.principals_factory",
    "kotti.register": "False",
    "kotti.register.group": "",
    "kotti.register.role": "",
    "kotti.request_factory": "kotti.request.Request",
    "kotti.reset_password_callback": "kotti.views.login.reset_password_callback",  # noqa
    "kotti.root_factory": "kotti.resources.default_get_root",
    "kotti.sanitizers": " ".join(
        [
            "xss_protection:kotti.sanitizers.xss_protection",
            "minimal_html:kotti.sanitizers.minimal_html",
            "no_html:kotti.sanitizers.no_html",
        ]
    ),
    "kotti.sanitize_on_write": " ".join(
        [
            "kotti.resources.Document.body:xss_protection",
            "kotti.resources.Content.title:no_html",
            "kotti.resources.Content.description:no_html",
        ]
    ),
    "kotti.search_content": "kotti.views.util.default_search_content",
    "kotti.session_factory": "kotti.beaker_session_factory",
    "kotti.static.edit_needed": "",  # BBB
    "kotti.static.view_needed": "",  # BBB
    "kotti.templates.api": "kotti.views.util.TemplateAPI",
    "kotti.time_format": "medium",
    "kotti.url_normalizer": "kotti.url_normalizer.url_normalizer",
    "kotti.url_normalizer.map_non_ascii_characters": True,
    "kotti.use_tables": "",
    "kotti.use_workflow": "kotti:workflow.zcml",
    "kotti.zcml_includes": " ".join([]),
    "pyramid.includes": "",
    "pyramid_deform.template_search_path": "kotti:templates/deform",
}

conf_dotted = {
    "kotti.authn_policy_factory",
    "kotti.authz_policy_factory",
    "kotti.available_types",
    "kotti.base_includes",
    "kotti.caching_policy_chooser",
    "kotti.configurators",
    "kotti.fanstatic.edit_needed",
    "kotti.fanstatic.view_needed",
    "kotti.login_success_callback",
    "kotti.modification_date_excludes",
    "kotti.populators",
    "kotti.principals_factory",
    "kotti.request_factory",
    "kotti.reset_password_callback",
    "kotti.root_factory",
    "kotti.search_content",
    "kotti.session_factory",
    "kotti.templates.api",
    "kotti.url_normalizer",
}


def get_version():
    return pkg_resources.require("Kotti")[0].version


def get_settings():
    return get_current_registry().settings


def _resolve_dotted(d, keys=conf_dotted):
    resolved = d.copy()

    for key in keys:
        value = resolved[key]
        if not isinstance(value, str):
            continue
        new_value = []
        for dottedname in value.split():
            new_value.append(DottedNameResolver().resolve(dottedname))
        resolved[key] = new_value

    return resolved


def main(global_config, **settings):
    # This function is a 'paste.app_factory' and returns a WSGI
    # application.

    from kotti.resources import initialize_sql

    config = base_configure(global_config, **settings)
    engine = engine_from_config(config.registry.settings)
    initialize_sql(engine)
    return config.make_wsgi_app()


def base_configure(global_config, **settings):
    # Resolve dotted names in settings, include plug-ins and create a
    # Configurator.

    from kotti.resources import get_root

    for key, value in conf_defaults.items():
        settings.setdefault(key, value)

    for key, value in settings.items():
        if key.startswith("kotti") and isinstance(value, bytes):
            settings[key] = value.decode("utf8")

    # Allow extending packages to change 'settings' w/ Python:
    k = "kotti.configurators"
    for func in _resolve_dotted(settings, keys=(k,))[k]:
        func(settings)

    settings = _resolve_dotted(settings)
    secret1 = settings["kotti.secret"]
    settings.setdefault("kotti.secret2", secret1)

    # We'll process ``pyramid_includes`` later by hand, to allow
    # overrides of configuration from ``kotti.base_includes``:
    pyramid_includes = settings.pop("pyramid.includes", "")

    config = Configurator(
        request_factory=settings["kotti.request_factory"][0], settings=settings
    )
    config.begin()

    config.hook_zca()
    config.include("pyramid_zcml")

    # Chameleon bindings were removed from Pyramid core since pyramid>=1.5a2
    config.include("pyramid_chameleon")

    config.registry.settings["pyramid.includes"] = pyramid_includes

    # Include modules listed in 'kotti.base_includes':
    for module in settings["kotti.base_includes"]:
        config.include(module)
    config.commit()

    # Modules in 'pyramid.includes' and 'kotti.zcml_includes' may
    # override 'kotti.base_includes':
    if pyramid_includes:
        for module in pyramid_includes.split():
            config.include(module)

    for name in settings["kotti.zcml_includes"].strip().split():
        config.load_zcml(name)

    config.commit()

    config._set_root_factory(get_root)

    return config


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    import kotti.views.util

    settings = config.get_settings()

    authentication_policy = settings["kotti.authn_policy_factory"][0](**settings)
    authorization_policy = settings["kotti.authz_policy_factory"][0](**settings)
    session_factory = settings["kotti.session_factory"][0](**settings)
    if authentication_policy:
        config.set_authentication_policy(authentication_policy)
    if authorization_policy:
        config.set_authorization_policy(authorization_policy)
    config.set_session_factory(session_factory)

    config.add_subscriber(kotti.views.util.add_renderer_globals, BeforeRender)

    for override in [
        a.strip() for a in settings["kotti.asset_overrides"].split() if a.strip()
    ]:
        config.override_asset(to_override="kotti", override_with=override)

    config.add_translation_dirs(f"{os.path.dirname(__file__)}/locale")
    # used to be
    # config.add_translation_dirs("kotti:locale")
    # which fails with recent pytest (works in non testing though)

    workflow = settings["kotti.use_workflow"]
    if workflow.lower() not in FALSE_VALUES:
        config.load_zcml(workflow)

    return config



import hashlib
import time
from typing import Dict
from typing import List
from typing import Optional
from urllib.parse import urlencode

from html2text import HTML2Text
from pyramid.renderers import render
from pyramid_mailer.mailer import Mailer
from pyramid_mailer.message import Message

from kotti import get_settings
from kotti.request import Request
from kotti.security import Principal

_inject_mailer = []


def get_mailer():
    # Consider that we may have persistent settings
    if _inject_mailer:
        return _inject_mailer[0]
    return Mailer.from_settings(get_settings())  # pragma: no cover


def make_token(user: Principal, seconds: Optional[float] = None) -> str:
    secret = get_settings()["kotti.secret2"]
    if seconds is None:
        seconds = time.time()
    token = f"{user.name}:{secret}:{seconds}"
    return "{}:{}".format(hashlib.sha224(token.encode("utf8")).hexdigest(), seconds)


def validate_token(user: Principal, token: str, valid_hrs: int = 24) -> bool:
    """
        >>> from kotti.testing import setUp, tearDown
        >>> ignore = setUp()
        >>> class User(object):
        ...     pass
        >>> daniel = User()
        >>> daniel.name = u'daniel'
        >>> alice = User()
        >>> alice.name = u'alice'
        >>> token = make_token(daniel)
        >>> validate_token(daniel, token)
        True
        >>> validate_token(alice, token)
        False
        >>> validate_token(daniel, 'foo')
        False
        >>> token = make_token(daniel, seconds=time.time() - 100000)
        >>> validate_token(daniel, token)
        False
        >>> validate_token(daniel, token, valid_hrs=48)
        True
        >>> tearDown()
    """
    try:
        seconds = float(token.split(":")[1])
    except (IndexError, ValueError):
        return False
    if (
        token == make_token(user, seconds)
        and time.time() - seconds < 60 * 60 * valid_hrs
    ):
        return True
    return False


def send_email(
    request: Request,
    recipients: List[str],
    template_name: str,
    template_vars: Optional[Dict[str, str]] = None,
) -> None:
    """ General email sender.

    :param request: current request.
    :type request: :class:`kotti.request.Request`

    :param recipients: list of email addresses. Each email should be a
                       string like: u'"John Doe" <joedoe@foo.com>'.
    :type recipients: list

    :param template_name: asset specification (e.g.
                          'mypackage:templates/email.pt')
    :type template_name: string

    :param template_vars: set of variables present on template.
    :type template_vars: dict
    """

    if template_vars is None:
        template_vars = {}

    text = render(template_name, template_vars, request)
    subject, htmlbody = text.strip().split("\n", 1)
    subject = subject.replace("Subject:", "", 1).strip()
    html2text = HTML2Text()
    html2text.body_width = 0
    textbody = html2text.handle(htmlbody).strip()

    message = Message(
        recipients=recipients, subject=subject, body=textbody, html=htmlbody
    )
    mailer = get_mailer()
    mailer.send(message)


def email_set_password(
    user: Principal,
    request: Request,
    template_name: Optional[str] = "kotti:templates/email-set-password.pt",  # noqa
    add_query: Optional[Dict[str, str]] = None,
) -> None:
    site_title = get_settings()["kotti.site_title"]
    token = make_token(user)
    user.confirm_token = token
    set_password_query = {"token": token, "email": user.email}
    if add_query:
        set_password_query.update(add_query)
    url = "{}/@@set-password?{}".format(
        request.application_url, urlencode(set_password_query)
    )
    variables = dict(user_title=user.title, site_title=site_title, url=url)
    recipients = [f'"{user.title}" <{user.email}>']  # XXX naive?
    send_email(request, recipients, template_name, variables)



import traceback

import transaction
from alembic import context
from zope.sqlalchemy import mark_changed

from kotti import DBSession
from kotti import metadata


def run_migrations_online():
    if DBSession.bind is None:
        raise ValueError("You must run Kotti's migration using the "
                         "'kotti-migrate' script and not through 'alembic' "
                         "directly.")

    transaction.begin()
    connection = DBSession.connection()

    context.configure(
        connection=connection,
        target_metadata=metadata,
        )

    try:
        context.run_migrations()
        mark_changed(DBSession())
    except:  # noqa: E722
        traceback.print_exc()
        transaction.abort()
    else:
        transaction.commit()
    finally:
        # connection.close()
        pass


try:  # Alembic's "if __name__ == '__main__'"
    offline_mode = context.is_offline_mode()
except (AttributeError, NameError):
    pass
else:
    if offline_mode:  # pragma: no cover
        raise ValueError("No support for Alembic's offline mode at this point. "
                         "You may want to write your own env.py script to use "
                         "'offline mode'.")
    run_migrations_online()



import warnings

from pytest import raises


def assert_deprecations(w, *msgs):

    assert len(w) == len(msgs)

    for i in range(len(w)):
        assert issubclass(w[i].category, DeprecationWarning)
        assert msgs[i] in str(w[i].message)


class TestToBeRemovedIn20:
    def test_type_info_action_links_deprecated(self, allwarnings):
        from kotti.resources import TypeInfo
        from kotti.util import LinkParent

        my_item = object()
        with warnings.catch_warnings(record=True) as wngs:
            # If there's a last LinkParent item, we'll assume that is
            # the action menu.
            TypeInfo(edit_links=[LinkParent("foo", [])], action_links=[my_item])
            assert wngs[0].category == DeprecationWarning

        with raises(ValueError):
            # If there's no last LinkParent item, we'll raise an
            # error, since we can't do anything useful with the link.
            TypeInfo(edit_links=[], action_links=[my_item])

    def test_object_after_delete(self, allwarnings):
        with warnings.catch_warnings(record=True) as w:
            from kotti.events import ObjectAfterDelete

            __ = ObjectAfterDelete  # pyflakes
            assert_deprecations(w, "Kotti 0.10")

    def test_translate_titles_deprecated(self, allwarnings):
        from kotti.views.edit import _translate_titles

        with warnings.catch_warnings(record=True) as w:
            info = [
                {"data": {"title": "_('Private')"}, "title": "_('Private')"},
                {"data": {"title": "_('Public')"}, "title": "_('Public')"},
            ]
            _translate_titles(info)
            assert_deprecations(w, "removed in Kotti 2.0.0")



import hashlib
from collections import defaultdict
from datetime import datetime
from urllib.parse import urlencode

from babel.dates import format_date
from babel.dates import format_datetime
from babel.dates import format_time
from babel.numbers import format_currency
from pyramid.decorator import reify
from pyramid.i18n import get_locale_name
from pyramid.interfaces import ILocation
from pyramid.location import inside
from pyramid.location import lineage
from pyramid.renderers import get_renderer
from pyramid.renderers import render
from pyramid.settings import asbool
from sqlalchemy import and_
from sqlalchemy import not_
from sqlalchemy import or_

from kotti import DBSession
from kotti import get_settings
from kotti.events import objectevent_listeners
from kotti.interfaces import INavigationRoot
from kotti.resources import Content
from kotti.resources import Document
from kotti.resources import Node
from kotti.resources import Tag
from kotti.resources import TagsToContents
from kotti.resources import get_root
from kotti.sanitizers import sanitize
from kotti.security import view_permitted
from kotti.util import TemplateStructure
from kotti.util import render_view
from kotti.views.site_setup import CONTROL_PANEL_LINKS
from kotti.views.slots import slot_events


class SettingHasValuePredicate:
    def __init__(self, val, config):
        self.name, self.value = val
        if not isinstance(self.value, bool):
            raise ValueError("Only boolean values supported")

    def text(self):
        return f"if_setting_has_value = {self.name} == {self.value}"

    phash = text

    def __call__(self, context, request):
        return asbool(request.registry.settings[self.name]) == self.value


class RootOnlyPredicate:
    def __init__(self, val, config):
        self.val = val

    def text(self):
        return f"root_only = {self.val}"

    phash = text

    def __call__(self, context, request):
        return (context is request.root) == self.val


def template_api(context, request, **kwargs):
    return get_settings()["kotti.templates.api"][0](context, request, **kwargs)


def add_renderer_globals(event):
    if event.get("renderer_name") != "json":
        request = event["request"]
        api = getattr(request, "template_api", None)
        if api is None and request is not None:
            api = template_api(event["context"], event["request"])
        event["api"] = api


class Slots:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __getattr__(self, key):
        for event_type in slot_events:
            if event_type.name == key:
                break
        else:
            raise AttributeError(key)
        value = []
        event = event_type(self.context, self.request)
        for snippet in objectevent_listeners(event):
            if snippet is not None:
                if isinstance(snippet, list):
                    value.extend(snippet)
                else:
                    value.append(snippet)
        setattr(self, key, value)
        return value


class TemplateAPI:
    """This implements the ``api`` object that's passed to all templates.

    Use dict-access as a shortcut to retrieve template macros from templates.
    """

    # Instead of overriding these, consider using the
    # ``kotti.overrides`` variable.
    BARE_MASTER = "kotti:templates/master-bare.pt"
    VIEW_MASTER = "kotti:templates/view/master.pt"
    EDIT_MASTER = "kotti:templates/edit/master.pt"
    SITE_SETUP_MASTER = "kotti:templates/site-setup/master.pt"

    body_css_class = ""

    def __init__(self, context, request, bare=None, **kwargs):
        self.context, self.request = context, request

        if getattr(request, "template_api", None) is None:
            request.template_api = self

        self.S = get_settings()
        if request.is_xhr and bare is None:
            bare = True  # use bare template that renders just the content area
        self.bare = bare
        self.slots = Slots(context, request)
        self.__dict__.update(kwargs)

    @staticmethod
    def is_location(context):
        """Does `context` implement :class:`pyramid.interfaces.ILocation`?

        :param context: The context.
        :type context: kotti.interfaces.INode
        :rtype: bool
        :returns: True if Is the context object implements
                  :class:`pyramid.interfaces.ILocation`.
        """
        return ILocation.providedBy(context)

    @reify
    def edit_needed(self):
        if "kotti.fanstatic.edit_needed" in self.S:
            return [r.need() for r in self.S["kotti.fanstatic.edit_needed"]]

    @reify
    def view_needed(self):
        if "kotti.fanstatic.view_needed" in self.S:
            return [r.need() for r in self.S["kotti.fanstatic.view_needed"]]

    def macro(self, asset_spec, macro_name="main"):
        if self.bare and asset_spec in (
            self.VIEW_MASTER,
            self.EDIT_MASTER,
            self.SITE_SETUP_MASTER,
        ):
            asset_spec = self.BARE_MASTER
        return get_renderer(asset_spec).implementation().macros[macro_name]

    @reify
    def site_title(self):
        """ The site title.

        :result: Value of the ``kotti.site_title`` setting (if specified) or
                 the root item's ``title`` attribute.
        :rtype: str
        """
        value = get_settings().get("kotti.site_title")
        if not value:
            value = self.root.title
        return value

    @reify
    def page_title(self):
        """
        Title for the current page as used in the ``<head>`` section of the
        default ``master.pt`` template.

        :result: '[Human readable view title ]``context.title`` -
                 :meth:`~TemplateAPI.site_title`''
        :rtype: str
        """

        view_title = self.request.view_name.replace("_", " ").title()
        if view_title:
            view_title += " "
        view_title += self.context.title
        return f"{view_title} - {self.site_title}"

    def url(self, context=None, *elements, **kwargs):
        """
        URL construction helper. Just a convenience wrapper for
        :func:`pyramid.request.resource_url` with the same signature.  If
        ``context`` is ``None`` the current context is passed to
        ``resource_url``.
        """

        if context is None:
            context = self.context
        if not ILocation.providedBy(context):
            return self.request.url
        return self.request.resource_url(context, *elements, **kwargs)

    @reify
    def root(self):
        """
        The site root.

        :result: The root object of the site.
        :rtype: :class:`kotti.resources.Node`
        """

        if ILocation.providedBy(self.context):
            return self.lineage[-1]
        else:
            return get_root()

    @reify
    def navigation_root(self):
        """
        The root node for the navigation.

        :result: Nearest node in the :meth:`lineage` that provides
                 :class:`kotti.interfaces.INavigationRoot` or :meth:`root` if
                 no node provides that interface.
        :rtype: :class:`kotti.resources.Node`
        """
        for o in self.lineage:
            if INavigationRoot.providedBy(o):
                return o
        return self.root

    @reify
    def lineage(self):
        """
        Lineage from current context to the root node.

        :result: List of nodes.
        :rtype: list of :class:`kotti.resources.Node`
        """
        return list(lineage(self.context))

    @reify
    def breadcrumbs(self):
        """
        List of nodes from the :meth:`navigation_root` to the context.

        :result: List of nodes.
        :rtype: list of :class:`kotti.resources.Node`
        """
        breadcrumbs = self.lineage
        if self.root != self.navigation_root:
            index = breadcrumbs.index(self.navigation_root)
            breadcrumbs = breadcrumbs[: index + 1]
        return reversed(breadcrumbs)

    def has_permission(self, permission, context=None):
        """ Convenience wrapper for :func:`pyramid.security.has_permission`
        with the same signature.  If ``context`` is ``None`` the current
        context is passed to ``has_permission``."""
        if context is None:
            context = self.context
        return self.request.has_permission(permission, context)

    def render_view(self, name="", context=None, request=None, secure=True, bare=True):
        if context is None:
            context = self.context
        if request is None:
            request = self.request

        before = self.bare
        try:
            self.bare = bare
            html = render_view(context, request, name, secure)
        finally:
            self.bare = before
        return TemplateStructure(html)

    def render_template(self, renderer, **kwargs):
        return TemplateStructure(render(renderer, kwargs, self.request))

    def list_children(self, context=None, permission="view"):

        if context is None:
            context = self.context

        if isinstance(context, Node):
            if permission is None:
                return context.children
            return context.children_with_permission(self.request, permission)

        return [
            c
            for c in getattr(context, "values", lambda: [])()
            if (not permission or self.request.has_permission(permission, c))
        ]

    inside = staticmethod(inside)

    def avatar_url(self, user=None, size="14", default_image="identicon"):
        if user is None:
            user = self.request.user
        email = user.email
        if not email:
            email = user.name
        h = hashlib.md5(email.encode("utf8")).hexdigest()
        query = {"default": default_image, "size": str(size)}
        url = "https://secure.gravatar.com/avatar/{}?{}".format(h, urlencode(query))
        return url

    @reify
    def locale_name(self):
        return get_locale_name(self.request)

    def format_date(self, d, fmt=None):
        if fmt is None:
            fmt = self.S["kotti.date_format"]
        return format_date(d, format=fmt, locale=self.locale_name)

    def format_datetime(self, dt, fmt=None):
        if fmt is None:
            fmt = self.S["kotti.datetime_format"]
        if not isinstance(dt, datetime):
            dt = datetime.fromtimestamp(dt)
        return format_datetime(dt, format=fmt, locale=self.locale_name)

    def format_time(self, t, fmt=None):
        if fmt is None:
            fmt = self.S["kotti.time_format"]
        return format_time(t, format=fmt, locale=self.locale_name)

    def format_currency(self, n, currency, fmt=None):
        return format_currency(n, currency, format=fmt, locale=self.locale_name)

    @staticmethod
    def get_type(name):
        for class_ in get_settings()["kotti.available_types"]:
            if class_.type_info.name == name:
                return class_

    def find_edit_view(self, item):
        view_name = self.request.view_name
        if not view_permitted(item, self.request, view_name):
            view_name = "edit"
        if not view_permitted(item, self.request, view_name):
            view_name = ""
        return view_name

    @reify
    def edit_links(self):
        if not hasattr(self.context, "type_info"):
            return []
        return [
            link
            for link in self.context.type_info.edit_links
            if link.visible(self.context, self.request)
        ]

    @reify
    def site_setup_links(self):
        return [link for link in CONTROL_PANEL_LINKS
                if link.visible(self.root, self.request)]

    @staticmethod
    def sanitize(html, sanitizer="default"):
        """ Convenience wrapper for :func:`kotti.sanitizers.sanitize`.

        :param html: HTML to be sanitized
        :type html: str

        :param sanitizer: name of the sanitizer to use.
        :type sanitizer: str

        :result: sanitized HTML
        :rtype: str
        """

        return sanitize(html, sanitizer)


class NodesTree:
    def __init__(self, node, request, item_mapping, item_to_children, permission):
        self._node = node
        self._request = request
        self._item_mapping = item_mapping
        self._item_to_children = item_to_children
        self._permission = permission

    @property
    def __parent__(self):
        if self.parent_id:
            return self._item_mapping[self.parent_id]

    @property
    def children(self):
        return [
            NodesTree(
                child,
                self._request,
                self._item_mapping,
                self._item_to_children,
                self._permission,
            )
            for child in self._item_to_children[self.id]
            if self._request.has_permission(self._permission, child)
        ]

    def _flatten(self, item):
        # noinspection PyProtectedMember
        yield item._node
        for ch in item.children:
            yield from self._flatten(ch)

    def tolist(self):
        return list(self._flatten(self))

    def __getattr__(self, key):
        return getattr(self._node, key)


def nodes_tree(request, context=None, permission="view"):
    item_mapping = {}
    item_to_children = defaultdict(lambda: [])
    for node in DBSession.query(Content).with_polymorphic(Content):
        item_mapping[node.id] = node
        if request.has_permission(permission, node):
            item_to_children[node.parent_id].append(node)

    for children in item_to_children.values():
        children.sort(key=lambda ch: ch.position)

    if context is None:
        node = item_to_children[None][0]
    else:
        node = context

    return NodesTree(node, request, item_mapping, item_to_children, permission)


def search_content(search_term, request=None):
    return get_settings()["kotti.search_content"][0](search_term, request)


def default_search_content(search_term, request=None):

    # noinspection PyUnresolvedReferences
    searchstring = f"%{search_term}%"

    # generic_filter can be applied to all Node (and subclassed) objects
    generic_filter = or_(
        Content.name.like(searchstring),
        Content.title.like(searchstring),
        Content.description.like(searchstring),
    )

    results = (
        DBSession.query(Content)
        .filter(generic_filter)
        .order_by(Content.title.asc())
        .all()
    )

    # specific result contain objects matching additional criteria
    # but must not match the generic criteria (because these objects
    # are already in the generic_results)
    document_results = DBSession.query(Document).filter(
        and_(Document.body.like(searchstring), not_(generic_filter))
    )

    for results_set in [content_with_tags([searchstring]), document_results.all()]:
        [results.append(c) for c in results_set if c not in results]

    result_dicts = []

    for result in results:
        if request.has_permission("view", result):
            result_dicts.append(
                dict(
                    name=result.name,
                    title=result.title,
                    description=result.description,
                    path=request.resource_path(result),
                )
            )

    return result_dicts


def content_with_tags(tag_terms):

    return (
        DBSession.query(Content)
        .join(TagsToContents)
        .join(Tag)
        .filter(or_(*[Tag.title.like(tag_term) for tag_term in tag_terms]))
        .all()
    )


def search_content_for_tags(tags, request=None):

    result_dicts = []

    for result in content_with_tags(tags):
        if request.has_permission("view", result):
            result_dicts.append(
                dict(
                    name=result.name,
                    title=result.title,
                    description=result.description,
                    path=request.resource_path(result),
                )
            )

    return result_dicts


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.add_view_predicate("root_only", RootOnlyPredicate)
    config.add_view_predicate("if_setting_has_value", SettingHasValuePredicate)



"""
Action views
"""
from pyramid.exceptions import Forbidden
from pyramid.httpexceptions import HTTPFound
from pyramid.url import resource_url
from pyramid.view import view_config
from pyramid.view import view_defaults

from kotti import DBSession
from kotti import get_settings
from kotti.fanstatic import contents_view_js
from kotti.interfaces import IContent
from kotti.resources import Node
from kotti.util import ActionButton
from kotti.util import _
from kotti.util import get_paste_items
from kotti.util import title_to_name
from kotti.views.edit import _state_info
from kotti.views.edit import _states
from kotti.views.form import EditFormView
from kotti.views.util import nodes_tree
from kotti.workflow import get_workflow


@view_defaults(permission="edit")
class NodeActions:
    """Actions related to content nodes."""

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.flash = self.request.session.flash

    def _selected_children(self, add_context=True):
        """ Get the selected children of the given context. These are either
        the selected nodes of the contents view or the context itself.

        :result: List with select children.
        :rtype: list
        """
        ids = self.request.session.pop("kotti.selected-children")
        if ids is None and add_context:
            ids = [self.context.id]
        return ids

    def _all_children(self, context, permission="view"):
        """ Recursively get all children of the given context.

        :result: List with all children of a given context.
        :rtype: list
        """
        tree = nodes_tree(self.request, context=context, permission=permission)
        return tree.tolist()[1:]

    def back(self, view=None):
        """ Redirect to the given view of the context, the referrer of the
        request or the default_view of the context.

        :rtype: pyramid.httpexceptions.HTTPFound
        """
        url = self.request.resource_url(self.context)
        if view is not None:
            url += view
        elif self.request.referrer:
            url = self.request.referrer
        return HTTPFound(location=url)

    @view_config(name="workflow-change", permission="state_change")
    def workflow_change(self):
        """ Handle workflow change requests from workflow dropdown.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        new_state = self.request.params["new_state"]
        wf = get_workflow(self.context)
        wf.transition_to_state(self.context, self.request, new_state)
        self.flash(EditFormView.success_message, "success")
        return self.back()

    @view_config(name="copy")
    def copy_node(self):
        """ Copy nodes view. Copy the current node or the selected nodes in the
        contents view and save the result in the session of the request.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        return self._do_copy_or_cut("copy", "copied")

    @view_config(name="cut")
    def cut_nodes(self):
        """ Cut nodes view. Cut the current node or the selected nodes in the
        contents view and save the result in the session of the request.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        return self._do_copy_or_cut("cut", "cut")

    def _do_copy_or_cut(self, action, action_title):
        ids = self._selected_children()
        self.request.session["kotti.paste"] = (ids, action)
        for id in ids:
            item = DBSession.query(Node).get(id)
            self.flash(
                _("${title} was %s." % action_title, mapping=dict(title=item.title)),
                "success",
            )
        if not self.request.is_xhr:
            return self.back()

    @view_config(name="paste")
    def paste_nodes(self):
        """ Paste nodes view.  Paste formerly copied or cutted nodes into the
        current context.  Note that a cutted node can not be pasted into itself.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        ids, action = self.request.session["kotti.paste"]
        for count, id in enumerate(ids):
            item = DBSession.query(Node).get(id)
            if item is not None:
                if action == "cut":
                    if not self.request.has_permission("edit", item):
                        raise Forbidden()
                    item.__parent__.children.remove(item)
                    item.name = title_to_name(item.name, blacklist=self.context.keys())
                    self.context[item.name] = item
                    if count is len(ids) - 1:
                        del self.request.session["kotti.paste"]
                elif action == "copy":
                    copy = item.copy()
                    name = copy.name
                    if not name:  # for root
                        name = copy.title
                    name = title_to_name(name, blacklist=self.context.keys())
                    copy.name = name
                    self.context[name] = copy
                self.flash(
                    _("${title} was pasted.", mapping=dict(title=item.title)), "success"
                )
            else:
                self.flash(_("Could not paste node. It no longer exists."), "error")
        DBSession.flush()
        if not self.request.is_xhr:
            return self.back()

    def move(self, move):
        """ Do the real work to move the selected nodes up or down. Called
        by the up and the down view.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        ids = self._selected_children()
        if move == 1:
            ids.reverse()
        for id in ids:
            child = DBSession.query(Node).get(id)
            index = self.context.children.index(child)
            self.context.children.pop(index)
            self.context.children.insert(index + move, child)
            self.flash(
                _("${title} was moved.", mapping=dict(title=child.title)), "success"
            )
        if not self.request.is_xhr:
            return self.back()

    @view_config(name="up")
    def up(self):
        """ Move up nodes view. Move the selected nodes up by 1 position
        and get back to the referrer of the request.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        return self.move(-1)

    @view_config(name="down")
    def down(self):
        """ Move down nodes view. Move the selected nodes down by 1 position
        and get back to the referrer of the request.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        return self.move(1)

    def set_visibility(self, show):
        """ Do the real work to set the visibility of nodes in the menu. Called
        by the show and the hide view.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        ids = self._selected_children()
        for id in ids:
            child = DBSession.query(Node).get(id)
            if child.in_navigation != show:
                child.in_navigation = show
                mapping = dict(title=child.title)
                if show:
                    mg = _(
                        "${title} is now visible in the navigation.", mapping=mapping
                    )
                else:
                    mg = _(
                        "${title} is no longer visible in the navigation.",
                        mapping=mapping,
                    )
                self.flash(mg, "success")
        if not self.request.is_xhr:
            return self.back()

    @view_config(name="show")
    def show(self):
        """ Show nodes view.  Switch the in_navigation attribute of selected
        nodes to ``True`` and get back to the referrer of the request.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        return self.set_visibility(True)

    @view_config(name="hide")
    def hide(self):
        """ Hide nodes view. Switch the in_navigation attribute of selected
        nodes to ``False`` and get back to the referrer of the request.

        :result: Redirect response to the referrer of the request.
        :rtype: pyramid.httpexceptions.HTTPFound
        """
        return self.set_visibility(False)

    @view_config(
        name="delete", permission="delete", renderer="kotti:templates/edit/delete.pt"
    )
    def delete_node(self):
        """ Delete node view. Renders either a view to delete the current node
        or handle the deletion of the current node and get back to the
        default view of the node.

        :result: Either a redirect response or a dictionary passed to the
                 template for rendering.
        :rtype: pyramid.httpexceptions.HTTPFound or dict
        """

        action = self.request.POST.get("delete")
        if action is not None:

            parent = self.context.__parent__

            if action == "delete":
                location = resource_url(parent, self.request)
                self.flash(
                    _("${title} was deleted.", mapping=dict(title=self.context.title)),
                    "success",
                )
                del parent[self.context.name]
            else:
                location = resource_url(self.context, self.request)

            return HTTPFound(location=location)

        return {}

    @view_config(
        name="delete_nodes",
        permission="delete",
        renderer="kotti:templates/edit/delete-nodes.pt",
    )
    def delete_nodes(self):
        """ Delete nodes view. Renders either a view to delete multiple nodes or
        delete the selected nodes and get back to the referrer of the request.

        :result: Either a redirect response or a dictionary passed to the
                 template for rendering.
        :rtype: pyramid.httpexceptions.HTTPFound or dict
        """
        if "delete_nodes" in self.request.POST:
            ids = self.request.POST.getall("children-to-delete")
            if not ids:
                self.flash(_("Nothing was deleted."), "info")
            for id in ids:
                item = DBSession.query(Node).get(id)
                self.flash(
                    _("${title} was deleted.", mapping=dict(title=item.title)),
                    "success",
                )
                del self.context[item.name]
            return self.back("@@contents")

        if "cancel" in self.request.POST:
            self.flash(_("No changes were made."), "info")
            return self.back("@@contents")

        ids = self._selected_children(add_context=False)
        items = []
        if ids is not None:
            items = (
                DBSession.query(Node)
                .filter(Node.id.in_(ids))
                .order_by(Node.position)
                .all()
            )
        return {"items": items, "states": _states(self.context, self.request)}

    @view_config(name="rename", renderer="kotti:templates/edit/rename.pt")
    def rename_node(self):
        """ Rename node view. Renders either a view to change the title and
        name for the current node or handle the changes and get back to the
        default view of the node.

        :result: Either a redirect response or a dictionary passed to the
                 template for rendering.
        :rtype: pyramid.httpexceptions.HTTPFound or dict
        """
        if "rename" in self.request.POST:
            name = self.request.POST["name"]
            title = self.request.POST["title"]
            if not name or not title:
                self.flash(_("Name and title are required."), "error")
            else:
                self.context.name = name.replace("/", "")
                self.context.title = title
                self.flash(_("Item was renamed."), "success")
                return self.back("")
        return {}

    @view_config(name="rename_nodes", renderer="kotti:templates/edit/rename-nodes.pt")
    def rename_nodes(self):
        """ Rename nodes view. Renders either a view to change the titles and
        names for multiple nodes or handle the changes and get back to the
        referrer of the request.

        :result: Either a redirect response or a dictionary passed to the
                 template for rendering.
        :rtype: pyramid.httpexceptions.HTTPFound or dict
        """
        if "rename_nodes" in self.request.POST:
            ids = self.request.POST.getall("children-to-rename")
            for id in ids:
                item = DBSession.query(Node).get(id)
                name = self.request.POST[id + "-name"]
                title = self.request.POST[id + "-title"]
                if not name or not title:
                    self.flash(_("Name and title are required."), "error")
                    location = resource_url(
                        self.context, self.request, "@@rename_nodes"
                    )
                    return HTTPFound(location=location)
                else:
                    item.name = title_to_name(name, blacklist=self.context.keys())
                    item.title = title
            self.flash(_("Your changes have been saved."), "success")
            return self.back("@@contents")

        if "cancel" in self.request.POST:
            self.flash(_("No changes were made."), "info")
            return self.back("@@contents")

        ids = self._selected_children(add_context=False)
        items = []
        if ids is not None:
            items = DBSession.query(Node).filter(Node.id.in_(ids)).all()
        return {"items": items}

    @view_config(name="change_state", renderer="kotti:templates/edit/change-state.pt")
    def change_state(self):
        """ Change state view. Renders either a view to handle workflow changes
        for multiple nodes or handle the selected workflow changes and get
        back to the referrer of the request.

        :result: Either a redirect response or a dictionary passed to the
                 template for rendering.
        :rtype: pyramid.httpexceptions.HTTPFound or dict
        """
        if "change_state" in self.request.POST:
            ids = self.request.POST.getall("children-to-change-state")
            to_state = self.request.POST.get("to-state", "no-change")
            include_children = self.request.POST.get("include-children")
            if to_state != "no-change":
                items = DBSession.query(Node).filter(Node.id.in_(ids)).all()
                for item in items:
                    wf = get_workflow(item)
                    if wf is not None:
                        wf.transition_to_state(item, self.request, to_state)
                    if include_children:
                        childs = self._all_children(item, permission="state_change")
                        for child in childs:
                            wf = get_workflow(child)
                            if wf is not None:
                                wf.transition_to_state(child, self.request, to_state)
                self.flash(_("Your changes have been saved."), "success")
            else:
                self.flash(_("No changes were made."), "info")
            return self.back("@@contents")

        if "cancel" in self.request.POST:
            self.flash(_("No changes were made."), "info")
            return self.back("@@contents")

        ids = self._selected_children(add_context=False)
        items = transitions = []
        if ids is not None:
            wf = get_workflow(self.context)
            if wf is not None:
                items = DBSession.query(Node).filter(Node.id.in_(ids)).all()
                for item in items:
                    trans_info = wf.get_transitions(item, self.request)
                    for tran_info in trans_info:
                        if tran_info not in transitions:
                            transitions.append(tran_info)
        return {
            "items": items,
            "states": _states(self.context, self.request),
            "transitions": transitions,
        }


def contents_buttons(context, request):
    """ Build the action buttons for the contents view based on the current
    state and the persmissions of the user.

    :result: List of ActionButtons.
    :rtype: list
    """
    buttons = []
    if get_paste_items(context, request):
        buttons.append(ActionButton("paste", title=_("Paste"), no_children=True))
    if context.children:
        buttons.append(ActionButton("copy", title=_("Copy")))
        buttons.append(ActionButton("cut", title=_("Cut")))
        buttons.append(
            ActionButton("rename_nodes", title=_("Rename"), css_class="btn btn-warning")
        )
        buttons.append(
            ActionButton("delete_nodes", title=_("Delete"), css_class="btn btn-danger")
        )
        if get_workflow(context) is not None:
            buttons.append(ActionButton("change_state", title=_("Change State")))
        buttons.append(ActionButton("up", title=_("Move up")))
        buttons.append(ActionButton("down", title=_("Move down")))
        buttons.append(ActionButton("show", title=_("Show")))
        buttons.append(ActionButton("hide", title=_("Hide")))
    return [button for button in buttons if button.permitted(context, request)]


@view_config(name="add-dropdown", renderer="kotti:templates/add-dropdown.pt")
def content_type_factories(context, request):
    """ Renders the drop down menu for Add button in editor bar.

    :result: Dictionary passed to the template for rendering.
    :rtype: pyramid.httpexceptions.HTTPFound or dict
    """
    all_types = get_settings()["kotti.available_types"]
    factories = []
    for factory in all_types:
        if factory.type_info.addable(context, request):
            factories.append(factory)
    return {"factories": factories}


@view_config(
    context=IContent,
    name="contents",
    permission="view",
    renderer="kotti:templates/edit/contents.pt",
)
def contents(context, request):
    """ Contents view. Renders either the contents view or handle the action
    button actions of the view.

    :result: Either a redirect response or a dictionary passed to the
             template for rendering.
    :rtype: pyramid.httpexceptions.HTTPFound or dict
    """
    contents_view_js.need()
    buttons = contents_buttons(context, request)
    for button in buttons:
        if button.name in request.POST:
            children = request.POST.getall("children")
            if not children and button.name != "paste":
                request.session.flash(
                    _("You have to select items to perform an action."), "info"
                )
                location = resource_url(context, request) + "@@contents"
                return HTTPFound(location=location)
            request.session["kotti.selected-children"] = children
            location = button.url(context, request)
            return HTTPFound(location, request=request)

    return {"children": context.children_with_permission(request), "buttons": buttons}


@view_config(
    name="move-child-position",
    permission="edit",
    request_method="POST",
    renderer="json",
)
def move_child_position(context, request):
    """ Move the child from one position to another.

    :param context: "Container" node in which the child changes its position.
    :type context: :class:kotti.resources.Node or descendant

    :param request: Current request (of method POST).  Must contain either
                    "from" and "to" params or a json_body that contain(s) the
                    0-based old (i.e. the current index of the child to be
                    moved) and new position (its new index) values.
    :type request:
    :result: JSON serializable object with a single attribute ("result") that is
             either "success" or "error".
    :rtype: dict
    """

    data = request.POST or request.json_body

    if ("from" in data) and ("to" in data):

        max_pos = len(context.children) - 1
        try:
            old_position = int(data["from"])
            new_position = int(data["to"])
            if not ((0 <= old_position <= max_pos) and (0 <= new_position <= max_pos)):
                raise ValueError
        except ValueError:
            return {"result": "error"}

        # sqlalchemy.ext.orderinglist takes care of the "right" sequence
        # numbers (immediately consecutive, starting with 0) for us.
        context.children.insert(new_position, context.children.pop(old_position))
        result = "success"
    else:
        result = "error"

    return {"result": result}


@view_config(
    name="workflow-dropdown",
    permission="view",
    renderer="kotti:templates/workflow-dropdown.pt",
)
def workflow(context, request):
    """ Renders the drop down menu for workflow actions.

    :result: Dictionary passed to the template for rendering.
    :rtype: dict
    """
    wf = get_workflow(context)
    if wf is not None:
        state_info = _state_info(context, request)
        curr_state = [i for i in state_info if i["current"]][0]
        trans_info = [
            trans
            for trans in wf.get_transitions(context, request)
            if request.has_permission(trans["permission"], context)
        ]

        return {
            "states": _states(context, request),
            "transitions": trans_info,
            "current_state": curr_state,
        }

    return {"current_state": None}


@view_config(
    name="actions-dropdown",
    permission="edit",
    renderer="kotti:templates/actions-dropdown.pt",
)
def actions(context, request):
    """ Renders the drop down menu for Actions button in editor bar.

    :result: Dictionary passed to the template for rendering.
    :rtype: dict
    """
    actions = []
    if hasattr(context, "type_info"):
        actions = [
            a for a in context.type_info.action_links if a.visible(context, request)
        ]
    return {"actions": actions}


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    import warnings

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        config.scan(__name__)
